#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>

#ifdef DICT_HW_MEASURE_TIMINGS
# include <sys/time.h>
#endif /* DICT_HW_MEASURE_TIMINGS */

#if defined(_WIN32) || defined(_WIN64)
# include "win/getopt.h"
#else 
# include <getopt.h>
#endif

#include <glib.h>

#include <dict_hw.h>

#include "helpers.h"

int matlab_mode = 0;

typedef enum {
    OPT_DEFAULT_FFT = 250,
    OPT_SINGLE_MODE = 251,
    OPT_MATLAB = 'm',
    OPT_HELP = 'h'
} Options;

static struct option long_options[] = {
    {"matlab",			no_argument, 0, OPT_MATLAB },
    {"default-fft",		no_argument, 0, OPT_DEFAULT_FFT },
    {"no-threads",		no_argument, 0, OPT_SINGLE_MODE },
    {"help",			no_argument, 0, OPT_HELP },
    { 0, 0, 0, 0 }
};

void Usage(int argc, char *argv[]) {
    printf(
" %s [options] [mode] <directory> [maximages]\n"
"  Modes:\n"
"	--matlab	- Matlab mode (transposed data ordering)\n"
"	--default-fft	- Do not optimize FFT size\n"
"	--no-threads	- Single GPU mode\n"
"	--help		- Help message\n"
"\n\n", argv[0]);
}

int readfile(char *name, float *x, int size) {
    int i;

    char first_line[16384];

    struct stat st;
    stat(name, &st);

    FILE *f = fopen(name, "r");

    if (!f) {
	printf("File %s is not found\n", name);
	g_free(name);
	exit(-1);
    }

    g_free(name);


    if (matlab_mode) {
	int len = 0;
        int columns = 0;
	int block_len;
	do {
	    fgets(first_line, 16384, f);
	    for (block_len = 0; first_line[block_len]; block_len++) {
		if (first_line[block_len] == '.') columns++;
	    }
	    len += block_len;
	} while (block_len == 16383);
    
	int lines = st.st_size / len;
    
	int pos = 0;
	rewind(f);
	for (i = 0; fscanf(f, "%f", x+pos) == 1; i++) {
	    if ((pos+1) == size) {
		size *= 2;
		x = (float*)realloc(x, size);
	    }
	
	    pos = ((i+1)%columns)*lines + ((i+1)/columns);
	}
    } else {
	for (i = 0; fscanf(f, "%f", x+i) == 1; i++) {
	    if ((i+1) == size) {
		size *= 2;
		x = (float*)realloc(x, size);
	    }
	}
    }

    fclose(f);

    return i;
}

int writefile(char *name, float *x, int img, int ncp) {
    FILE *f = fopen(name, "w");
    if (!f) {
	printf("Can't write file %s\n", name);
	g_free(name);
	exit(-1);
    }
    g_free(name);

    for (int j = 0; j < ncp ; j++) {    
	for (int i = 0; i < img ; i++) {
	    fprintf(f, "% 9.7le\t", x[i*ncp + j]);
	}
	fprintf(f,"\n");
    }
    
    fclose(f);
}

int compare(const void *a, const void*b) {
    return strcmp(*(char**)a, *(char**)b);
}

int main(int argc, char *argv[]) {
    int i;
    int err;
    int flags;
    
    int maximg, allocimg;
    
    int size = 32768;
    int xsize, ysize;
    float *x;
    float *y;

    char **images;

#ifdef DICT_HW_MEASURE_TIMINGS
    int time[8];
    struct timeval tv1, tv2, tv3;

    memset(time, 0, sizeof(time));
    gettimeofday(&tv1, NULL);
#endif /* DICT_HW_MEASURE_TIMINGS */

    int default_fft = 0;
    int single_mode = 0;
    
    int option_index = 0;
    unsigned char c;
    while ((c = getopt_long(argc, argv, "yhe:f:l:t:c:", long_options, &option_index)) != (unsigned char)-1) {
	switch (c) {
	    case 0:
	    break;
	    case OPT_MATLAB:
		matlab_mode = 1;
	    break;
	    case OPT_DEFAULT_FFT:
		default_fft = 1;
	    break;
	    case OPT_SINGLE_MODE:
		single_mode = 1;
	    break;
	    case OPT_HELP:
		Usage(argc, argv);
		exit(0);		
	    break;
	    default:
		Usage(argc, argv);
		exit(0);
	}
    }

    if (optind > argc - 1) {
	Usage(argc, argv);
	exit(0);
    }
    
    x = (float*)malloc(size * sizeof(float));
    y = (float*)malloc(size * sizeof(float));

    gchar *name = g_strdup_printf("%s/data/grid_x.dat", argv[optind]);
    xsize = readfile(name, x, size);

    name = g_strdup_printf("%s/data/grid_y.dat", argv[optind]);
    ysize = readfile(name, y, size);
    
    if (xsize != ysize) {
	printf("Amount of X and Y coordinates of control points does not match\n");
	exit(-1);
    }
    
    //xsize = ysize = 1024;

    if (optind < (argc - 1)) maximg = atoi(argv[optind+1]);
    else maximg = 0;
    
    images = (char**)malloc(size * sizeof(char*));
    
    name = g_strdup_printf("%s/images", argv[optind]);
    GDir *dir = g_dir_open(name, 0, NULL);

    if (!dir) {
	printf("Directory %s is not found\n", name);
	g_free(name);
	exit(-1);
    }
    
    g_free(name);

    GRegex *regex = g_regex_new("(tiff?|bmp|jpe?g|png)", G_REGEX_CASELESS, (GRegexMatchFlags)0, NULL);
    if (!regex) {
	printf("Internal error. Failed to compiler regular expression\n");
	exit(-1);
    }
    
    
    const gchar *file = g_dir_read_name(dir);
    for (i = 0;file;file = g_dir_read_name(dir)) {
	if (!g_regex_match(regex, file, (GRegexMatchFlags)0, NULL)) continue;
	
	if (i == size) {
	    size *= 2;
	    images = (char**)realloc(images, size);
	}
	images[i++] = g_strdup(file);
/*	if (!strstr(images[i-1], "tif")) {
	    puts(images[i-1]);
	    exit(-1);
	}*/
    }    
    g_dir_close(dir);
    g_regex_unref(regex);

    allocimg = i;
    if ((!maximg)||(maximg > i)) maximg = i;
    
    qsort(images, allocimg, sizeof(char*), (int (*)(const void*, const void*))compare);
    
    if (maximg < 2) {
	printf("At least two images are required\n");
	exit(-1);
    }

    int ncp = xsize;
    
    DICTContext ps = dictCreateContext();
    if (!ps) {
	printf("Context intialization is failed, no CUDA-devices...\n");
	exit(-1);
    }

    flags = DICT_FLAGS_DEFAULT;
    if (matlab_mode) flags |= DICT_FLAGS_MATLAB_MODE;
    if (default_fft) flags |= DICT_FLAGS_FIXED_FFT_SIZE;
    if (single_mode) flags |= DICT_FLAGS_SINGLE_THREAD;

    err = dictSetup(ps, ncp, 15, 1000, flags);
    
    
    if (err) {
	printf("Setup is failed, error code %i\n", err);
	exit(-1);
    }

    float *res;
    int step;
    
    if (matlab_mode) {
	res = (float*) malloc(2 * ncp * (maximg - 1) * sizeof(float));
	step = ncp * (maximg - 1);
    } else {
	res = (float*) malloc(2 * ncp * sizeof(float));
	step = ncp;
	dictSetPointsBuffer(ps, res, res + ncp);
    }

    dictSetTemplatePoints(ps, x, y);

    name = g_strdup_printf("%s/images/%s", argv[optind], images[0]);
    if (!g_file_test(name, G_FILE_TEST_IS_REGULAR)) {
	g_free(name);
	printf("%s is not a regular file\n", images[0]);
	exit(-1);
    }
    err = dictLoadTemplateImageFile(ps, name);
    g_free(name);
    if (err) {
	printf("failed to load template file %s\n", images[0]);
	exit(-1);
    }
    

    dictSetCurrentPoints(ps, x, y);

#ifdef DICT_HW_MEASURE_TIMINGS
    gettimeofday(&tv2, NULL);
    time[0] = (tv2.tv_sec - tv1.tv_sec) * 1000000 + (tv2.tv_usec - tv1.tv_usec);
#endif /* DICT_HW_MEASURE_TIMINGS */


#ifdef DICT_HW_MEASURE_TIMINGS
    gettimeofday(&tv3, NULL);
#endif /* DICT_HW_MEASURE_TIMINGS */

    FILE *fx, *fy;
    if (!matlab_mode) {
	name = g_strdup_printf("%s/data/validx.dat", argv[optind]);
	fx = fopen(name, "w");
	g_free(name);
	name = g_strdup_printf("%s/data/validy.dat", argv[optind]);
	fy = fopen(name, "w");
	g_free(name);
	
	if ((!fx)||(!fy)) {
	    printf("Can't create output files...\n");
	    exit(-1);
	}
    }

#ifdef DICT_HW_MEASURE_TIMINGS
    gettimeofday(&tv2, NULL);
    time[1] += (tv2.tv_sec - tv3.tv_sec) * 1000000 + (tv2.tv_usec - tv3.tv_usec);
#endif /* DICT_HW_MEASURE_TIMINGS */


    for (i = 1; i < maximg; i++) {
#ifndef DICT_HW_MEASURE_TIMINGS
	printf("%s\n", images[i]);
#endif /* !DICT_HW_MEASURE_TIMINGS */
	
	name = g_strdup_printf("%s/images/%s", argv[optind], images[i]);
	if (!g_file_test(name, G_FILE_TEST_IS_REGULAR)) {
	    g_free(name);
	    printf("%s is not a regular file\n", images[0]);
	    exit(-1);
	}

#ifdef DICT_HW_MEASURE_TIMINGS
        gettimeofday(&tv3, NULL);
#endif /* DICT_HW_MEASURE_TIMINGS */
	err = dictLoadImageFile(ps, name);
#ifdef DICT_HW_MEASURE_TIMINGS
	gettimeofday(&tv2, NULL);
	time[5] += (tv2.tv_sec - tv3.tv_sec) * 1000000 + (tv2.tv_usec - tv3.tv_usec);
#endif /* DICT_HW_MEASURE_TIMINGS */
	g_free(name);

	if (err) {
	    printf("failed to load image %s\n", images[0]);
	    exit(-1);
	}

	if (matlab_mode) {
	    float *out = res + ncp * (i - 1);
	    dictGetCurrentPoints(ps, out, out + step);
	} else {
	    dictCompute(ps);

#ifdef DICT_HW_MEASURE_TIMINGS
	    gettimeofday(&tv3, NULL);
#endif /* DICT_HW_MEASURE_TIMINGS */

	    for (int j = 0; j < ncp; j++) {
		fprintf(fx, "% 9.7le\t", res[j]);
	    }
	    fprintf(fx,"\n");
	    
	    float *out = res + step;
	    for (int j = 0; j < ncp; j++) {
		fprintf(fy, "% 9.7le\t", out[j]);
	    }
	    fprintf(fy,"\n");

#ifdef DICT_HW_MEASURE_TIMINGS
	    gettimeofday(&tv2, NULL);
	    time[1] += (tv2.tv_sec - tv3.tv_sec) * 1000000 + (tv2.tv_usec - tv3.tv_usec);
#endif /* DICT_HW_MEASURE_TIMINGS */
	}
    }

#ifdef DICT_HW_MEASURE_TIMINGS
    gettimeofday(&tv3, NULL);
#endif /* DICT_HW_MEASURE_TIMINGS */
    
    if (matlab_mode) {
	name = g_strdup_printf("%s/data/validx.dat", argv[optind]);
	writefile(name, res, maximg - 1, ncp);

	name = g_strdup_printf("%s/data/validy.dat", argv[optind]);
	writefile(name, res+step, maximg - 1, ncp);
    } else {
	fclose(fx);
	fclose(fy);
    }

#ifdef DICT_HW_MEASURE_TIMINGS
    gettimeofday(&tv2, NULL);
    time[1] += (tv2.tv_sec - tv3.tv_sec) * 1000000 + (tv2.tv_usec - tv3.tv_usec);
    time[2] = (tv2.tv_sec - tv1.tv_sec) * 1000000 + (tv2.tv_usec - tv1.tv_usec) - time[0];
#endif /* DICT_HW_MEASURE_TIMINGS */

    for (i = 0; i < allocimg; i++) {
	g_free(images[i]);
    }
    
    dictDestroyContext(ps);
    
    free(res);
    free(images);        
    free(y);
    free(x);

#ifdef DICT_HW_MEASURE_TIMINGS
    gettimeofday(&tv3, NULL);
    time[3] = (tv3.tv_sec - tv1.tv_sec) * 1000000 + (tv3.tv_usec - tv1.tv_usec);
    time[4] = (tv3.tv_sec - tv2.tv_sec) * 1000000 + (tv3.tv_usec - tv2.tv_usec);
    reportMessage("");
    reportMessage("Global timings");
    reportMessage("--------------");
    print_timing("Overall", time[3]);
    print_timing("Initialization", time[0]);
    print_timing("Cleanup", time[4]);
    print_timing("Processing/Full", time[2], maximg - 1);
    print_timing("Processing/Computations", time[2] - time[1], maximg - 1);
    print_timing("Processing/Pushing Data", time[5], maximg - 1);
    print_timing("Processing/Waiting Result", time[2] - time[1] - time[5], maximg - 1);
    print_timing("Processing/Writeout", time[1], maximg - 1);
    reportMessage("");
#endif /* DICT_HW_MEASURE_TIMINGS */

}

